#!/usr/bin/env python

# Copyright 2015, 2017, 2019-2020 National Research Foundation (SARAO)
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import configparser
import glob
import os
import os.path
import subprocess

from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext

import pybind11


def find_version():
    # Cannot simply import it, since that tries to import spead2 as well, which
    # isn't built yet.
    globals_ = {}
    with open(os.path.join(os.path.dirname(__file__), 'src', 'spead2', '_version.py')) as f:
        code = f.read()
    exec(code, globals_)
    return globals_['__version__']


class BuildExt(build_ext):
    user_options = build_ext.user_options + [
        ('coverage', None,
         'build with GCC --coverage option'),
        ('split-debug=', None,
         'write debug symbols to a separate directory')
    ]
    boolean_options = build_ext.boolean_options + ['coverage']

    def initialize_options(self):
        build_ext.initialize_options(self)
        # setuptools bug causes these to be lost during reinitialization by
        # ./setup.py develop
        if not hasattr(self, 'coverage'):
            self.coverage = None
        if not hasattr(self, 'split_debug'):
            self.split_debug = None

    def run(self):
        self.mkpath(self.build_temp)
        subprocess.check_call(os.path.abspath('configure'), cwd=self.build_temp)
        config = configparser.ConfigParser()
        config.read(os.path.join(self.build_temp, 'python-build.cfg'))
        for extension in self.extensions:
            extension.extra_compile_args.extend(config['compiler']['CFLAGS'].split())
            extension.extra_link_args.extend(config['compiler']['LIBS'].split())
            if self.coverage:
                extension.extra_compile_args.extend(['-g', '--coverage'])
                extension.libraries.extend(['gcov'])
            if self.split_debug:
                extension.extra_compile_args.extend(['-g'])
            extension.include_dirs.insert(0, os.path.join(self.build_temp, 'include'))
        # distutils uses old-style classes, so no super
        build_ext.run(self)

    def build_extensions(self):
        # Stop GCC complaining about -Wstrict-prototypes in C++ code
        try:
            self.compiler.compiler_so.remove('-Wstrict-prototypes')
        except ValueError:
            pass
        build_ext.build_extensions(self)

    def build_extension(self, ext):
        ext_path = self.get_ext_fullpath(ext.name)
        if self.split_debug:
            # If the base class decides to skip the link, we'll end up
            # constructing the .debug file from the already-stripped version of
            # the library, and it won't have any symbols. So force the link by
            # removing the output.
            try:
                os.remove(ext_path)
            except OSError:
                pass
        build_ext.build_extension(self, ext)
        if self.split_debug:
            os.makedirs(self.split_debug, exist_ok=True)
            basename = os.path.basename(ext_path)
            debug_path = os.path.join(self.split_debug, basename + '.debug')
            self.spawn(['objcopy', '--only-keep-debug', '--', ext_path, debug_path])
            self.spawn(['strip', '--strip-debug', '--', ext_path])
            old_cwd = os.getcwd()
            # See the documentation for --add-gnu-debuglink for why it needs to be
            # run from the directory containing the file.
            ext_path_abs = os.path.abspath(ext_path)
            try:
                os.chdir(self.split_debug)
                self.spawn(['objcopy', '--add-gnu-debuglink=' + os.path.basename(debug_path),
                            '--', ext_path_abs])
            finally:
                os.chdir(old_cwd)
            self.spawn(['chmod', '-x', '--', debug_path])


# Can't actually install on readthedocs.org because we can't compile,
# but we need setup.py to still be successful to make the doc build work.
rtd = os.environ.get('READTHEDOCS') == 'True'

if not rtd:
    if not os.path.exists(os.path.join(os.path.dirname(__file__), 'configure')):
        raise SystemExit("configure not found. Either download a release " +
                         "from https://pypi.python.org/pypi/spead2 or run " +
                         "./bootstrap.sh if not using a release.")

    extensions = [
        Extension(
            '_spead2',
            sources=(glob.glob('src/common_*.cpp') +
                     glob.glob('src/recv_*.cpp') +
                     glob.glob('src/send_*.cpp') +
                     glob.glob('src/py_*.cpp')),
            depends=glob.glob('include/spead2/*.h'),
            language='c++',
            include_dirs=['include', pybind11.get_include()],
            extra_compile_args=['-std=c++11', '-g0', '-fvisibility=hidden'])
    ]
else:
    extensions = []

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme_file:
    readme = readme_file.read()

setup(
    author='Bruce Merry',
    author_email='bmerry@ska.ac.za',
    name='spead2',
    version=find_version(),
    description='High-performance SPEAD implementation',
    long_description=readme,
    url='https://github.com/ska-sa/spead2',
    license='LGPLv3+',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries',
        'Topic :: System :: Networking'],
    ext_package='spead2',
    ext_modules=extensions,
    cmdclass={'build_ext': BuildExt},
    install_requires=[
        'numpy>=1.9.2'
    ],
    tests_require=[
        'netifaces',
        'pytest',
        'pytest-asyncio',
        'pytest-timeout'
    ],
    python_requires='>=3.6',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={'': ['py.typed', '*.pyi']},
    entry_points={
        'console_scripts': [
            'spead2_send.py = spead2.tools.send_asyncio:main',
            'spead2_recv.py = spead2.tools.recv_asyncio:main',
            'spead2_bench.py = spead2.tools.bench_asyncio:main'
        ]
    }
)



# Copyright 2020 National Research Foundation (SARAO)
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Command-line processing utilities for the tools in this module.
"""

import spead2.recv
import spead2.send


_HAVE_IBV = hasattr(spead2.recv.Stream, 'add_udp_ibv_reader')


def parse_endpoint(endpoint):
    if ':' in endpoint:
        host, port = endpoint.rsplit(':', 1)
        port = int(port)
    else:
        host = ''
        port = int(endpoint)
    return host, port


class _Options:
    """Base class for classes holding command-line arguments.

    The derived class must have attributes corresponding to the command-line
    arguments (with the same name).

    Parameters
    ----------
    name_map : Optional[Mapping[str, Optional[str]]]
        If specified, provides for renaming and suppressing command-line
        options. If ``name_map['foo_bar'] == 'abc_xyz'`` then the end user will
        pass ``--abc-xyz`` instead of ``--foo-bar``. Setting an element to
        ``None`` will suppress the option.
    """

    def __init__(self, name_map=None):
        self._name_map = name_map or {}

    def _add_argument(self, parser, name, *args, **kwargs):
        """Add an argument to a parser.

        This is called by subclasses to add an argument to a parser.

        It functions like :py:meth:`argparse.ArgumentParser.add_argument`, except
        that:

        - Instead of the flag, one must provide the name of the class
          attribute. The flag is computed by converting underscores to dashes and
          prepending ``--`` (after applying the name map given to the
          constructor).
        - No `default` should be provided. The value currently stored in the object
          is used as the default.
        """
        assert 'dest' not in kwargs
        new_name = self._name_map.get(name, name)
        if new_name is not None:
            flag = '--' + new_name.replace('_', '-')
            parser.add_argument(flag, *args, dest=new_name, default=getattr(self, name), **kwargs)

    def _extract_args(self, namespace):
        """Update the object attributes from parsed arguments.

        After parsing the arguments, the subclass should call this to reflect
        the arguments into the object.
        """
        for name in self.__dict__:
            if name.startswith('_'):
                continue
            mapped_name = self._name_map.get(name, name)
            if mapped_name is not None:
                setattr(self, name, getattr(namespace, mapped_name))


class ProtocolOptions(_Options):
    """Options that the sender and receiver need to agree on."""

    def __init__(self, name_map=None) -> None:
        super().__init__(name_map)
        self.tcp = False
        self.pyspead = False

    def add_arguments(self, parser):
        self._add_argument(parser, 'tcp', action='store_true',
                           help='Use TCP instead of UDP')
        self._add_argument(parser, 'pyspead', action='store_true',
                           help='Be bug-compatible with PySPEAD')

    def notify(self, parser, namespace):
        self._extract_args(namespace)


class SharedOptions(_Options):
    """Options common to senders and receiver.

    Unlike :class:`ProtocolOptions`, these need not be the same at the sender
    and receiver.
    """

    def __init__(self, protocol, name_map=None) -> None:
        super().__init__(name_map)
        self._protocol = protocol
        self.buffer = None            # Default depends on protocol
        self.bind = None
        if _HAVE_IBV:
            # These must be set conditionally, because _extract_args requires
            # a command-line argument corresponding to every attribute.
            self.ibv = False
            self.ibv_vector = 0
        self.threads = 1
        self.affinity = None

    def add_arguments(self, parser):
        self._add_argument(parser, 'buffer', type=int, help='Socket buffer size')
        self._add_argument(parser, 'bind', type=str, help='Interface address')
        if _HAVE_IBV:
            self._add_argument(parser, 'ibv', action='store_true', help='Use ibverbs [no]')
            self._add_argument(parser, 'ibv_vector', type=int, metavar='N',
                               help='Completion vector, or -1 to use polling [%(default)s]')
            self._add_argument(parser, 'ibv_max_poll', type=int,
                               help='Maximum number of times to poll in a row [%(default)s]')
        self._add_argument(parser, 'affinity', type=spead2.parse_range_list,
                           help='List of CPUs to pin threads to [no affinity]')
        self._add_argument(parser, 'threads', type=int,
                           help='Number of worker threads [%(default)s]')

    def make_thread_pool(self):
        if self.affinity:
            spead2.ThreadPool.set_affinity(self.affinity[0])
            return spead2.ThreadPool(self.threads, self.affinity[1:] + self.affinity[:1])
        else:
            return spead2.ThreadPool(self.threads)


class ReceiverOptions(SharedOptions):
    """Options for receivers."""

    def __init__(self, protocol, name_map=None) -> None:
        super().__init__(protocol, name_map)
        self.memcpy_nt = False
        self.concurrent_heaps = spead2.recv.StreamConfig.DEFAULT_MAX_HEAPS
        self.ring_heaps = spead2.recv.RingStreamConfig.DEFAULT_HEAPS
        self.mem_pool = False
        self.mem_lower = 16384
        self.mem_upper = 32 * 1024**2
        self.mem_max_free = 12
        self.mem_initial = 8
        self.packet = None
        if _HAVE_IBV:
            self.ibv_max_poll = spead2.recv.UdpIbvConfig.DEFAULT_MAX_POLL

    def add_arguments(self, parser):
        self._add_argument(parser, 'memcpy_nt', action='store_true',
                           help='Use non-temporal memcpy')
        self._add_argument(parser, 'concurrent_heaps', type=int,
                           help='Maximum number of in-flight heaps [%(default)s]')
        self._add_argument(parser, 'ring_heaps', type=int,
                           help='Ring buffer capacity in heaps [%(default)s]')
        self._add_argument(parser, 'mem_pool', action='store_true', help='Use a memory pool')
        self._add_argument(parser, 'mem_lower', type=int,
                           help='Minimum allocation which will use the memory pool [%(default)s]')
        self._add_argument(parser, 'mem_upper', type=int,
                           help='Maximum allocation which will use the memory pool [%(default)s]')
        self._add_argument(parser, 'mem_max_free', type=int,
                           help='Maximum free memory buffers [%(default)s]')
        self._add_argument(parser, 'mem_initial', type=int,
                           help='Initial free memory buffers [%(default)s]')
        self._add_argument(parser, 'packet', type=int, help='Maximum packet size to accept')
        super().add_arguments(parser)

    def notify(self, parser, namespace):
        self._extract_args(namespace)
        if _HAVE_IBV:
            if self.ibv and not self.bind:
                parser.error('--ibv requires --bind')
            if self._protocol.tcp and self.ibv:
                parser.error('--ibv and --tcp are incompatible')

        if self.buffer is None:
            if self._protocol.tcp:
                self.buffer = spead2.recv.asyncio.Stream.DEFAULT_TCP_BUFFER_SIZE
            elif _HAVE_IBV and self.ibv:
                self.buffer = spead2.recv.UdpIbvConfig.DEFAULT_BUFFER_SIZE
            else:
                self.buffer = spead2.recv.asyncio.Stream.DEFAULT_UDP_BUFFER_SIZE

        if self.packet is None:
            if self._protocol.tcp:
                self.packet = spead2.recv.asyncio.Stream.DEFAULT_TCP_MAX_SIZE
            elif _HAVE_IBV and self.ibv:
                self.packet = spead2.recv.UdpIbvConfig.DEFAULT_MAX_SIZE
            else:
                self.packet = spead2.recv.asyncio.Stream.DEFAULT_UDP_MAX_SIZE

    def make_stream_config(self):
        config = spead2.recv.StreamConfig()
        config.max_heaps = self.concurrent_heaps
        config.bug_compat = spead2.BUG_COMPAT_PYSPEAD_0_5_2 if self._protocol.pyspead else 0
        if self.mem_pool:
            config.memory_allocator = spead2.MemoryPool(self.mem_lower, self.mem_upper,
                                                        self.mem_max_free, self.mem_initial)
        if self.memcpy_nt:
            config.memcpy = spead2.MEMCPY_NONTEMPORAL
        return config

    def make_ring_stream_config(self):
        return spead2.recv.RingStreamConfig(heaps=self.ring_heaps)

    def add_readers(self, stream, endpoints, *, allow_pcap=False):
        ibv_endpoints = []
        for source in endpoints:
            try:
                host, port = parse_endpoint(source)
            except ValueError:
                if not allow_pcap:
                    raise
                try:
                    stream.add_udp_pcap_file_reader(source)
                except AttributeError:
                    raise RuntimeError('spead2 was compiled without pcap support') from None
            else:
                if self._protocol.tcp:
                    stream.add_tcp_reader(port, self.packet, self.buffer, host)
                elif _HAVE_IBV and self.ibv:
                    ibv_endpoints.append((host, port))
                elif self.bind:
                    stream.add_udp_reader(host, port, self.packet, self.buffer, self.bind)
                else:
                    stream.add_udp_reader(port, self.packet, self.buffer, host)
        if ibv_endpoints:
            stream.add_udp_ibv_reader(
                spead2.recv.UdpIbvConfig(
                    endpoints=ibv_endpoints,
                    interface_address=self.bind or '',
                    max_size=self.packet,
                    buffer_size=self.buffer,
                    comp_vector=self.ibv_vector,
                    max_poll=self.ibv_max_poll))


class SenderOptions(SharedOptions):
    """Options for senders."""

    def __init__(self, protocol, name_map=None):
        super().__init__(protocol, name_map)
        self.addr_bits = spead2.Flavour().heap_address_bits
        self.packet = spead2.send.StreamConfig.DEFAULT_MAX_PACKET_SIZE
        self.burst = spead2.send.StreamConfig.DEFAULT_BURST_SIZE
        self.burst_rate_ratio = spead2.send.StreamConfig.DEFAULT_BURST_RATE_RATIO
        self.max_heaps = spead2.send.StreamConfig.DEFAULT_MAX_HEAPS
        self.rate_method = spead2.send.StreamConfig.DEFAULT_RATE_METHOD
        self.rate = 0.0
        self.ttl = None
        if _HAVE_IBV:
            self.ibv_max_poll = spead2.send.UdpIbvConfig.DEFAULT_MAX_POLL

    def add_arguments(self, parser):
        def parse_rate_method(value):
            try:
                return spead2.send.RateMethod.__members__[value.upper()]
            except KeyError as exc:
                raise ValueError from exc

        rate_method_names = list(spead2.send.RateMethod.__members__)

        self._add_argument(parser, 'addr_bits', type=int,
                           help='Heap address bits [%(default)s]')
        self._add_argument(parser, 'packet', type=int,
                           help='Maximum packet size to send [%(default)s]')
        self._add_argument(parser, 'burst', metavar='BYTES', type=int,
                           help='Burst size [%(default)s]')
        self._add_argument(parser, 'burst_rate_ratio', metavar='RATIO', type=float,
                           help='Hard rate limit, relative to --rate [%(default)s]')
        self._add_argument(parser, 'max_heaps', metavar='HEAPS', type=int,
                           help='Maximum heaps in flight [%(default)s]')
        self._add_argument(parser, 'rate_method', metavar='METHOD', type=parse_rate_method,
                           help=f'Method for rate limiting ({"/".join(rate_method_names)})')
        self._add_argument(parser, 'rate', metavar='Gb/s', type=float,
                           help='Transmission rate bound [no limit]')
        self._add_argument(parser, 'ttl', type=int, help='TTL for multicast target')
        super().add_arguments(parser)

    def notify(self, parser, namespace):
        self._extract_args(namespace)
        if _HAVE_IBV:
            if self.ibv and not self.bind:
                parser.error('--ibv requires --bind')
            if self._protocol.tcp and self.ibv:
                parser.error('--ibv and --tcp are incompatible')
        if self.buffer is None:
            if self._protocol.tcp:
                self.buffer = spead2.send.asyncio.TcpStream.DEFAULT_BUFFER_SIZE
            elif _HAVE_IBV and self.ibv:
                self.buffer = spead2.send.UdpIbvConfig.DEFAULT_BUFFER_SIZE
            else:
                self.buffer = spead2.send.asyncio.UdpStream.DEFAULT_BUFFER_SIZE

    def make_flavour(self):
        bug_compat = spead2.BUG_COMPAT_PYSPEAD_0_5_2 if self._protocol.pyspead else 0
        return spead2.Flavour(4, 64, self.addr_bits, bug_compat)

    def make_stream_config(self):
        return spead2.send.StreamConfig(
            max_packet_size=self.packet,
            rate=self.rate * (1e9 / 8),
            burst_size=self.burst,
            burst_rate_ratio=self.burst_rate_ratio,
            max_heaps=self.max_heaps,
            rate_method=self.rate_method)

    async def make_stream(self, thread_pool, endpoints, memory_regions):
        config = self.make_stream_config()
        if self._protocol.tcp:
            return await spead2.send.asyncio.TcpStream.connect(
                thread_pool, endpoints, config, self.buffer, self.bind or '')
        elif _HAVE_IBV and self.ibv:
            return spead2.send.asyncio.UdpIbvStream(
                thread_pool,
                config,
                spead2.send.UdpIbvConfig(
                    endpoints=endpoints,
                    interface_address=self.bind or '',
                    buffer_size=self.buffer,
                    ttl=self.ttl or 1,
                    comp_vector=self.ibv_vector,
                    max_poll=self.ibv_max_poll,
                    memory_regions=memory_regions
                )
            )
        else:
            kwargs = {}
            if self.ttl is not None:
                kwargs['ttl'] = self.ttl
            if self.bind:
                kwargs['interface_address'] = self.bind
            return spead2.send.asyncio.UdpStream(
                thread_pool, endpoints, config, self.buffer, **kwargs)



# Copyright 2019-2020 National Research Foundation (SARAO)
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Iterator, Any, List, Tuple, Sequence, Union, Text, Optional, ClassVar, overload
import socket

import spead2
from spead2 import _EndpointList

class RawItem:
    @property
    def id(self) -> int: ...
    @property
    def is_immediate(self) -> bool: ...
    @property
    def immediate_value(self) -> int: ...

class _HeapBase:
    @property
    def cnt(self) -> int: ...
    @property
    def flavour(self) -> spead2.Flavour: ...
    def get_items(self) -> List[RawItem]: ...
    def is_start_of_stream(self) -> bool: ...
    def is_end_of_stream(self) -> bool: ...

class Heap(_HeapBase):
    def get_descriptors(self) -> List[spead2.RawDescriptor]: ...

class IncompleteHeap(_HeapBase):
    @property
    def heap_length(self) -> int: ...
    @property
    def received_length(self) -> int: ...
    @property
    def payload_ranges(self) -> List[Tuple[int, int]]: ...

class StreamStats:
    heaps: int
    incomplete_heaps_evicted: int
    incomplete_Heaps_flushed: int
    packets: int
    batches: int
    worker_blocked: int
    max_batch: int
    single_packet_heaps: int
    search_dist: int
    def __add__(self, other: StreamStats) -> StreamStats: ...
    def __iadd__(self, other: StreamStats) -> None: ...

class StreamConfig:
    DEFAULT_MAX_HEAPS: ClassVar[int] = ...
    max_heaps: int
    bug_compat: int
    memcpy: int
    memory_allocator: spead2.MemoryAllocator
    stop_on_stop_item: bool
    allow_unsized_heaps: bool
    allow_out_of_order: bool
    def __init__(self, *, max_heaps: int = ..., bug_compat: int = ...,
                 memcpy: int = ..., memory_allocator: spead2.MemoryAllocator = ...,
                 stop_on_stop_item: bool = ..., allow_unsized_heaps: bool = ...,
                 allow_out_of_order: bool = ...) -> None: ...

class RingStreamConfig:
    DEFAULT_HEAPS: ClassVar[int]
    heaps: int
    contiguous_only: bool
    incomplete_keep_payload_ranges: bool
    def __init__(self, *, heaps: int = ..., contiguous_only: bool = ...,
                 incomplete_keep_payload_ranges: bool = ...) -> None: ...

class UdpIbvConfig:
    DEFAULT_BUFFER_SIZE: ClassVar[int]
    DEFAULT_MAX_SIZE: ClassVar[int]
    DEFAULT_MAX_POLL: ClassVar[int]

    endpoints: _EndpointList
    interface_address: str
    buffer_size: int
    max_size: int
    comp_vector: int
    max_poll: int

    def __init__(self, *, endpoints: _EndpointList = ..., interface_address: str = ...,
                 buffer_size: int = ..., max_size: int = ..., comp_vector: int = ...,
                 max_poll: int = ...) -> None: ...

class Ringbuffer:
    def size(self) -> int: ...
    def capacity(self) -> int: ...

# We make a dummy _Stream base class because mypy objects to the async stream
# type overloading get as a coroutine.
class _Stream:
    DEFAULT_UDP_IBV_MAX_SIZE: ClassVar[int]
    DEFAULT_UDP_IBV_BUFFER_SIZE: ClassVar[int]
    DEFAULT_UDP_IBV_MAX_POLL: ClassVar[int]
    DEFAULT_UDP_MAX_SIZE: ClassVar[int]
    DEFAULT_UDP_BUFFER_SIZE: ClassVar[int]
    DEFAULT_TCP_MAX_SIZE: ClassVar[int]
    DEFAULT_TCP_BUFFER_SIZE: ClassVar[int]

    def __init__(self, thread_pool: spead2.ThreadPool, config: StreamConfig = ...,
                 ring_config: RingStreamConfig = ...) -> None: ...
    def __iter__(self) -> Iterator[Heap]: ...
    def get_nowait(self) -> Heap: ...
    def add_buffer_reader(self, buffer: Any) -> None: ...
    @overload
    def add_udp_reader(self, port: int, max_size: int = ..., buffer_size: int = ...,
                       bind_hostname: str = ...) -> None: ...
    @overload
    def add_udp_reader(self, socket: socket.socket, max_size: int = ...) -> None: ...
    @overload
    def add_udp_reader(self, multicast_group: str, port: int, max_size: int = ...,
                       buffer_size: int = ..., interface_address: str = ...) -> None: ...
    @overload
    def add_udp_reader(self, multicast_group: str, port: int, max_size: int = ...,
                       buffer_size: int = ..., interface_index: int = ...) -> None: ...
    @overload
    def add_tcp_reader(self, port: int, max_size: int = ..., buffer_size: int = ...,
                       bind_hostname: str = ...) -> None: ...
    @overload
    def add_tcp_reader(self, acceptor: socket.socket, max_size: int = ...) -> None: ...
    @overload
    def add_udp_ibv_reader(self, multicast_group: str, port: int,
                           interface_address: str,
                           max_size: int = ..., buffer_size: int = ...,
                           comp_vector: int = ..., max_poll: int = ...) -> None: ...
    @overload
    def add_udp_ibv_reader(self, endpoints: Sequence[Tuple[str, int]],
                           interface_address: str,
                           max_size: int = ..., buffer_size: int = ...,
                           comp_vector: int = ..., max_poll: int = ...) -> None: ...
    @overload
    def add_udp_ibv_reader(self, config: UdpIbvConfig) -> None: ...
    def add_inproc_reader(self, queue: spead2.InprocQueue) -> None: ...
    def stop(self) -> None: ...
    @property
    def fd(self) -> int: ...
    @property
    def stats(self) -> StreamStats: ...
    @property
    def ringbuffer(self) -> Ringbuffer: ...
    @property
    def config(self) -> StreamConfig: ...
    @property
    def ring_config(self) -> RingStreamConfig: ...

class Stream(_Stream):
    def get(self) -> Heap: ...
